@@ -7,7 +7,7 @@
 // ======================================================================
 // Basic Service Elements EN 300 196-1 D.6
 
-int ParseBasicService(struct asn1_parm *pc, u_char * p, u_char * end, int *basicService)
+int ParseBasicService(struct asn1_parm *pc, u_char * p, u_char * end, __u32 *basicService)
 {
 	return ParseEnum(pc, p, end, basicService);
 }
